from __future__ import print_function
import os
import argparse
import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision
from torch.autograd import Variable
import torch.optim as optim
from torchvision import datasets, transforms
from models.wideresnet import *
from models.resnetcifar import ResNet18_cifar
from autoattack.autoattack import AutoAttack
import numpy as np
os.environ["CUDA_VISIBLE_DEVICES"] = "2,3"
parser = argparse.ArgumentParser(description='PyTorch CIFAR PGD Attack Evaluation')
parser.add_argument('--test-batch-size', type=int, default=512, metavar='N',
                    help='input batch size for testing (default: 200)')
parser.add_argument('--no-cuda', action='store_true', default=False,
                    help='disables CUDA training')
parser.add_argument('--epsilon', default=0.031,
                    help='perturbation')
parser.add_argument('--num-steps', default=20, type=int,
                    help='perturb number of steps')
parser.add_argument('--step-size', default=0.003,
                    help='perturb step size')
parser.add_argument('--random',
                    default=True,
                    help='random initialization for PGD')
parser.add_argument('--model-path',
                    default='./model-cifar100_umap_lbgat_trades0-WideResNet_nodetach_w_o/model-res-epoch94.pt',
                    help='model for white-box attack evaluation')
parser.add_argument('--source-model-path',
                    default='./checkpoints/model_cifar_wrn.pt',
                    help='source model for black-box attack evaluation')
parser.add_argument('--target-model-path',
                    default='./checkpoints/model_cifar_wrn.pt',
                    help='target model for black-box attack evaluation')
parser.add_argument('--white-box-attack', default=True,
                    help='whether perform white-box attack')
parser.add_argument('--mark', default='cifar100_umap_lbgat_trades0_Wresnet_test', type=str,
                    help='log file name')
parser.add_argument('--widen_factor', default=10, type=int,
                    help='widen_factor for wideresnet')
parser.add_argument('--num_classes', default=100, type=int,
                    help='cifar10 or cifar100')
parser.add_argument('--dataparallel', default=True, type=bool,
                    help='whether model is trained with dataparallel')

args = parser.parse_args()

# settings
use_cuda = not args.no_cuda and torch.cuda.is_available()
device = torch.device("cuda" if use_cuda else "cpu")
kwargs = {'num_workers': 1, 'pin_memory': True} if use_cuda else {}

# set up data loader
transform_test = transforms.Compose([transforms.ToTensor(), ])

if args.num_classes == 100:
    testset = torchvision.datasets.CIFAR100(root='./data/cifar100', train=False, download=True,
                                            transform=transform_test)
    test_loader = torch.utils.data.DataLoader(testset, batch_size=args.test_batch_size, shuffle=False, **kwargs)
else:
    testset = torchvision.datasets.CIFAR10(root='./data/cifar10', train=False, download=True,
                                           transform=transform_test)
    test_loader = torch.utils.data.DataLoader(testset, batch_size=args.test_batch_size, shuffle=False, **kwargs)

def CW_loss(x, y):
    x_sorted, ind_sorted = x.sort(dim=1)
    ind = (ind_sorted[:, -1] == y).float()

    loss_value = -(x[np.arange(x.shape[0]), y] - x_sorted[:, -2] * ind - x_sorted[:, -1] * (1. - ind))
    return loss_value.mean()
def _pgd_whitebox_AA(model,
                  X,
                  y,
                  adversary,
                  epsilon=args.epsilon,
                  num_steps=args.num_steps,
                  step_size=args.step_size):
    out = model(X)
    err = (out.data.max(1)[1] != y.data).float().sum()

    X_pgd = adversary.run_standard_evaluation(X, y, bs=X.size(0))
    err_pgd = (model(X_pgd).data.max(1)[1] != y.data).float().sum()
    return err, err_pgd


def eval_adv_test_whitebox_AA(model, device, test_loader, adverary):
    """
    evaluate model by white-box attack
    """
    model.eval()
    robust_err_total = 0
    natural_err_total = 0

    for data, target in test_loader:
        data, target = data.to(device), target.to(device)
        # pgd attack
        X, y = Variable(data, requires_grad=True), Variable(target)
        err_natural, err_robust = _pgd_whitebox_AA(model, X, y, adverary)
        robust_err_total += err_robust
        natural_err_total += err_natural

    open("Logs/"+args.mark,"a+").write("robust_err_total: "+str(robust_err_total)+"\n")
    open("Logs/"+args.mark,"a+").write("natural_err_total: "+str(natural_err_total)+"\n")
def _pgd_whitebox(model,
                  X,
                  y,cw=False,
                  epsilon=args.epsilon,
                  num_steps=args.num_steps,
                  step_size=args.step_size):
    out = model(X)
    err = (out.data.max(1)[1] != y.data).float().sum()
    X_pgd = Variable(X.data, requires_grad=True)
    if args.random:
        random_noise = torch.FloatTensor(*X_pgd.shape).uniform_(-epsilon, epsilon).to(device)
        X_pgd = Variable(X_pgd.data + random_noise, requires_grad=True)

    for _ in range(num_steps):
        opt = optim.SGD([X_pgd], lr=1e-3)
        opt.zero_grad()

        with torch.enable_grad():
            if cw :
                loss = CW_loss(model(X_pgd), y)
            else:
                loss = nn.CrossEntropyLoss()(model(X_pgd), y)
        loss.backward()
        eta = step_size * X_pgd.grad.data.sign()
        X_pgd = Variable(X_pgd.data + eta, requires_grad=True)
        eta = torch.clamp(X_pgd.data - X.data, -epsilon, epsilon)
        X_pgd = Variable(X.data + eta, requires_grad=True)
        X_pgd = Variable(torch.clamp(X_pgd, 0, 1.0), requires_grad=True)
    err_pgd = (model(X_pgd).data.max(1)[1] != y.data).float().sum()
    # print('err pgd (white-box): ', err_pgd)
    return err, err_pgd


def eval_adv_test_whitebox(model, device, test_loader,cw=False):
    """
    evaluate model by white-box attack
    """
    model.eval()
    robust_err_total = 0
    natural_err_total = 0

    for idx, (data, target) in enumerate(test_loader):
        # print("{}/{}".format(idx, len(test_loader)))
        data, target = data.to(device), target.to(device)
        # pgd attack
        X, y = Variable(data, requires_grad=True), Variable(target)
        err_natural, err_robust = _pgd_whitebox(model, X, y,cw=cw)
        robust_err_total += err_robust
        natural_err_total += err_natural
    print('natural_acc_total: ', 10000-natural_err_total)
    print('robust_acc_total: ',10000- robust_err_total)
    open("Logs/"+args.mark,"a+").write("robust_acc_total: "+str(10000-robust_err_total)+"\n")
    open("Logs/"+args.mark,"a+").write("natural_acc_total: "+str(10000-natural_err_total)+"\n")


def main():
    if args.white_box_attack:
        # white-box attack
        args.mark = args.mark + '.log'
        open("Logs/" + args.mark, "a+").write('pgd white-box attack\n')

        model = WideResNet(num_classes=args.num_classes, widen_factor=args.widen_factor)
        # model = ResNet18_cifar(num_classes = 10)
        if args.dataparallel:
            model = nn.DataParallel(model).to(device)
        else:
            model = model.to(device)
        # for i in range(25):
        i=0
        model_path = args.model_path
        print(model_path)
        open("Logs/" + args.mark, "a+").write(model_path+'\n')

        model.load_state_dict(torch.load(model_path))
        adversary = AutoAttack(model, norm='Linf', eps=args.epsilon, version='standard', log_path = "Logs/"+args.mark)
        adversary.seed = 0
        open("Logs/" + args.mark, "a+").write('PGD-20' + '\n')
        eval_adv_test_whitebox(model, device, test_loader,cw=False)
        print('CW-20:')
        open("Logs/" + args.mark, "a+").write('CW-20' + '\n')
        eval_adv_test_whitebox(model, device, test_loader, cw=True)
        print('AA : ')
        model.eval()
        ### Evaluate AutoAttack ###
        l = [x for (x, y) in test_loader]
        x_test = torch.cat(l, 0)
        l = [y for (x, y) in test_loader]
        y_test = torch.cat(l, 0)
        epsilon = 8 / 255.

        # adversary = AutoAttack(model, norm=norm, eps=epsilon, log_path=os.path.join(args.save_dir, 'log_file.txt'),
        #                        version='standard')

        # run attack and save images

            # adv_complete = adversary.run_standard_evaluation(x_test, y_test, bs=1000)
        # adversary = AutoAttack(model, norm='Linf', eps=epsilon, version='standard',log_path=os.path.join(args.mark, 'log_file.txt'))
        with torch.no_grad():
            X_adv = adversary.run_standard_evaluation(x_test, y_test, bs=128)

        # eval_adv_test_whitebox_AA(model, device, test_loader, adversary)



if __name__ == '__main__':
    main()
